var debug;
var delim = "__delim__";


var App = Backbone.Model.extend({
});

var File = Backbone.Model.extend({
    url: "/file",
    defaults: function(){
        return {
            id: _.uniqueId("File_"),
            name : "",
            timestamp : new Date(),
            content: "",
            object: "",
            type: ""
        };
    },
    initialize: function(){
        this.url = this.url + "/" + this.get("name");
    },
    parse: function(obj){
        if(_.isObject(obj.object) === false){
            obj.object = JSON.parse(obj.object);
        }
        return obj;
    }
});

var FileList = Backbone.Collection.extend({
    url: "/filelist",
    model: File,
});

var Nodedef = Backbone.Model.extend({
    defaults: function(){
        return   {
            category: "",
            description: "",
            inputs: [],
            outputs: [],
            name: "",
            parameters: []
        };
    }
});

var _Nodedefs = Backbone.Collection.extend({
    model: Nodedef,
    url: "/nodedefs",
    initialize: function(){
        this.listenTo(Sheets, 'add', this.addDynamicNode);
        this.on('addDynamicIO', this.addDynamicIO);
        this.on('delDynamicIO', this.delDynamicIO);
        this.on("fetch", this.fetchAndAdd);
    },
    fetchAndAdd: function(url){
        var tmp = new _Nodedefs();
        var that = this;
        tmp.fetch({url: url, success:
            function(def){def.each(function(d){that.add(d);});}});
    },
    addDynamicNode: function(sheet){
        if(sheet.get("name") === "MAIN") return;
        this.add(new Nodedef({
            category: "Dynamic",
            name: sheet.get("name"),
            description : "Dynamic node"
        }));
    },
    addDynamicIO: function(role){
        if(role.sheet === "MAIN") return;
        var def = this.find(function(d){
            return role.sheet === d.get("name");});
        if(role.role === "input"){
            def.get("inputs").push({
                description: role.description,
                name: role.name,
                object_type: role.object_type
            });
            this.trigger("change", "addDynamicIO", role.sheet, "inputs", def.get("inputs"));
        } else if (role.role === "output"){
            def.get("outputs").push({
                description: role.description,
                name: role.name,
                object_type: role.object_type
            });
            this.trigger("change", "addDynamicIO", role.sheet, "outputs", def.get("outputs"));
        }
    },
    delDynamicIO: function(role){
        if(role.main === "MAIN") return;
        var def = this.find(function(d){
            var name = role.nodeId.split("_")[1];
            return name === d.name;});
    }
});


var Node = Backbone.Model.extend({
});

var Nodes = Backbone.Collection.extend({
    model: Node,
    initialize: function(){
        this.listenTo(Nodedefs, "change", this.applyChanges);
        this.on("addNode", this.addNode);
        this.on("reset", function(col, opts){
            _.each(opts.previousModels, function(model){
                model.trigger("remove");
            });
        });
        this.on("addNodeObj", this.addNodeObj);
    },
    applyChanges: function(event, nodename, target, newvalue){
        if(event === "addDynamicIO"){
            this.each(function(node){
                if(node.get("name") === nodename){
                    node.set(target, newvalue);
                    node.trigger("change");
                }
            });
        }
        console.log(this.toJSON());
        console.log("changed:", event, target, nodename, newvalue);
    },
    addNodeObj: function(node, sheetname){
        var def = Nodedefs.find(function(def){
            return def.get("name") === node.get("name");});
        var index = this.filter(function(n){
            return n.get("name") === def.get("name");
        }).length + 1;
        // node.set("index", index);
        // node.set("id", _.template("node_<%= name %>_<%= index %>",
                 // node.toJSON()));
        // console.log(def.toJSON());
        node.set("inputs", JSON.parse(JSON.stringify(def.get("inputs"))));
        node.set("outputs", JSON.parse(JSON.stringify(def.get("outputs"))));

        this.add(node);
    },
    addNode: function(nodedef, sheetname){
        var node = new Node(nodedef.toJSON());
        var index = this.filter(function(n){
            return n.get("name") === nodedef.get("name");
        }).length + 1;

        node.set("index", index);
        node.set("id", _.template("node_<%= name %>_<%= index %>",
                 node.toJSON()));
        node.set("pos", [0, 0]);
        node.set("sheet", sheetname);

        this.add(node);
    }
});

var Sheet = Backbone.Model.extend({
    defaults: function(){
        return {
            name: "",
            type: "",
            nodes: new Nodes(),
            links: new Links(),
            roles: new TerminalRoles(),
            isActive: false
        };
    },
    initialize: function(){
        this.on("addNode", this.addNode);
        this.on("addPartialLink", this.addPartialLink);
        this.on("deletePartialLink", this.deletePartialLink);
        this.on("updatePartialLink", this.updatePartialLink);
        this.on("addTerminalRole", this.addTerminalRole);
        this.listenTo(this.get("nodes"), "change", function(){
            this.get("links").trigger("change");});

    },
    addTerminalRole: function(type, nodeId, termName, rolename){
        var term = this.get("nodes").find(function(n){return n.get("id") === nodeId;});
        if(type === "input"){
            term = _.find(term.get("inputs"),  function(t){return t.name === termName;});
        }
        else if(type === "output"){
            term = _.find(term.get("outputs"), function(t){return t.name === termName;});
        }
        else if(type === "condition"){
            term = _.find(term.get("outputs"), function(t){return t.name === termName;});
        }

        var exist = this.get("roles").find(function(r){
            return r.get("nodeId") === nodeId && r.get("termName") === termName;});
        if(exist){
            this.get("roles").remove(exist);
        }
        this.get("roles").add(new TerminalRole({
            description: term.description,
            name: rolename,
            nodeId: nodeId,
            object_type: term.type,
            role: type,
            sheet: this.get("name"),
            termName: termName
        }));
    },
    addNode: function(nodedef){
        this.get("nodes").trigger("addNode", nodedef, this.get("name"));
    },
    addPartialLink: function(link){
        this.get("links").trigger("addPartialLink", link, this.get("name"));
    },
    deletePartialLink: function(link){
        this.get("links").trigger("deletePartialLink", link, this.get("name"));
    },
    updatePartialLink: function(event){
        this.get("links").trigger("updatePartialLink", event, this.get("name"));
    }
});

var Sheets = Backbone.Collection.extend({
    model: Sheet,
    initialize: function(){
        this.on("reset", this.resetall);
        this.on("replace", function(sheetsobj){
            this.reset();
            var sheets = _.map(JSON.parse(JSON.stringify(sheetsobj)), function(s){
                delete s.nodes;
                delete s.links;
                delete s.roles;
                var sheet = new Sheet(s);

                this.add(sheet);
                return sheet;
            }, this);

            _.each(_.zip(sheetsobj, sheets), function(tmp){
                var sheetobj = tmp[0];
                var sheet = tmp[1];

                _.each(sheetobj.nodes, function(n){
                    sheet.get("nodes").trigger("addNodeObj",
                        new Node(n), sheet.get("name"))});
                _.each(sheetobj.roles, function(r){
                    sheet.get("roles").add(new TerminalRole(r))});
                _.each(sheetobj.links, function(l){
                    sheet.get("links").trigger("addLinkObj",
                        new Link(l), sheet.get("name"));});
            }, this);
        }, this);
        this.on("activate", function(name){
            this.active = name;
            var active = this.find(function(sheet){
                sheet.set("isActive", false);
                return sheet.get("name") === this.active;
            }, this);
            active.set("isActive", true);
        });
        this.on("addNodeToActiveSheet", function(nodedef){
            var active = this.find(function(sheet){
                return sheet.get("name") === this.active;
            }, this);
            active.trigger("addNode", nodedef);
        });
        this.on("addPartialLinkToActiveSheet", function(link){
            var active = this.find(function(sheet){
                return sheet.get("name") === this.active;
            }, this);
            active.trigger("addPartialLink", link);
        });
        this.on("connectPartialLink", function(link){
            var active = this.find(function(sheet){
                return sheet.get("name") === this.active;
            }, this);
            var plink = active.get("links").find(function(l){return l.get("isPartial");});

            // assume input
            console.log(link.isPartial);
            if(link.isPartial === "from"){
                delete link.fromTermName;
                delete link.fromNodeId;
                delete link.fromPos;
            }
            else{
                delete link.toTermName;
                delete link.toNodeId;
                delete link.toPos;
            }
            plink.set(_.extend(link, {isPartial: false}), {validate: true});

        });

        this.on("addTerminalRoleToSheets", function(type, termid, sheetid, rolename){
            var nodeId = termid.split(delim)[0];
            var termName = termid.split(delim)[1];
            var sheetname = sheetid.split("_sheet")[0];

            this.find(function(sheet){
                return sheet.get("name") === sheetname;
            }).trigger("addTerminalRole", type, nodeId, termName, rolename);
        });
    },
    resetall: function(col, opts){
        _.each(opts.previousModels, function(model){
            model.trigger("remove");
        });
    }
});

var Link = Backbone.Model.extend({
    defaults: function(){
        return {
            fromNodeId: "",
            toNodeId: "",
            fromTermId: "",
            toTermId:"",
            isPartial: false,
            sheet: ""
        }
    },
    initialize: function (){
        this.on("change", this.updateId);
        this.on("invalid", this.remove);
        this.updateId();
    },
    updateId: function(){
        this.set("fromTermId", this.get("fromNodeId") + delim + this.get("fromTermName"));
        this.set("toTermId", this.get("toNodeId") + delim + this.get("toTermName"));
    },
    validate: function(attrs){
        if(attrs.fromNodeId === attrs.toNodeId){
            console.log("Error:" + "You cannot connect same nodes");
            return "You cannot connect same nodes";
        }
    }
});
var Links = Backbone.Collection.extend({
    model: Link,
    initialize: function(){
        this.on("addPartialLink", this.addPartialLink);
        this.on("deletePartialLink", this.deletePartialLink);
        this.on("updatePartialLink", this.updatePartialLink);
        this.on("reset", this.resetall);
        this.on("addLinkObj", this.addLinkObj);
    },
    addLinkObj: function(link, sheetname){
        this.add(link);
    },
    resetall: function(col, opts){
        _.each(opts.previousModels, function(model){
            model.trigger("remove");
        });
    },
    addPartialLink: function(link, sheetname){
        link.set("sheet", sheetname);
        this.add(link);
    },
    deletePartialLink: function(link){
        // this.find(this, function(l){return l.get("isPartial")}).remove();
        this.remove(this.find(function(l){return l.get("isPartial") !== false}));
    },
    updatePartialLink: function(pos){
        var partial = this.find(function(l){return l.get("isPartial") !== false;});
        if(partial){
            console.log(partial.get("isPartial"), pos);
            if(partial.get("isPartial") === "from")
                partial.set("fromPos", pos);
            else
                partial.set("toPos", pos);
        }
    }
});


var TerminalRole = Backbone.Model.extend({
    defaults: {
        description: "",
        name: "",
        nodeId: "",
        object_type: "",
        role: "",
        sheet: "",
        termName: ""
    }
});

var TerminalRoles = Backbone.Collection.extend({
    model: TerminalRole,
    initialize: function(){
        this.on("remove", this.removeFromDef);
        this.on("add", this.addToDef);
    },
    removeFromDef: function(role){
        Nodedefs.trigger("delDynamicIO", role.toJSON());
    },
    addToDef: function(role){
        Nodedefs.trigger("addDynamicIO", role.toJSON());
    }
});

